; -*- fundamental -*-
;:
; Firmware for ABC80 SecureDigital controller
;:
; This needs a CPU mostly because ABC wants 256-byte sectors, and
; SecureDigital has 512-byte sectors, at least if one wants to have
; SDHC compatibility.
;:
; We mimic a "real" microcontroller-based design in the sense that we
; have separate ROM and RAM, the RAM being uninitalized.
; Total memory: 1K ROM, 2K RAM
; RAM:
; 8000-81FF -> controller data and stack
; 8200-83FF -> space for one SD sector
; 8400-87FF -> space for four ABC sectors (RAMBUFs)
;

	; -----------------------------------------------------------
	;  Data area
	; -----------------------------------------------------------
sysinfo:	equ 0x8000		; IY points here
command:	equ 0x8000		; Four byte controller command
sd_sect:	equ 0x8004		; LBA of SD sector in SD buffer
sd_need:	equ 0x8008		; SD sect needed (temp for load_buffer)
abc_buf:	equ 0x800C		; Controller buffer addressed (0-3)
		; Used as a word address, so 800D is used.
init_done:	equ 0x800E		; Initialization done flag

		; SD card argument buffers
zero_arg:	equ 0x8010		; All zero argument
sect_arg:	equ 0x8014		; Sector number as argument
card_ocr:	equ 0x8018		; Bit 30 is the SDHC flag
init_arg:	equ 0x801c		; Normally zero but modified by sd_init
clear_size:	equ 0x20
stack:		equ 0x8200		; Base of stack

ram_size:	equ 2048

sd_buf:		equ 0x8200
rambuf:		equ 0x8400

reset:					; Controller reset
	di
	im 1				; Interrupt = RST 38h
	ld sp,stack			; Reset stack
	jr reset_1
rst8:	jp sd_cmd			; rst8 = call sd_cmd
reset_1:
	ld hl,sysinfo
	push hl
	pop iy
	ld de,sysinfo+1
	ld bc,clear_size-1		; Size of data area
	ld (hl),l			; Zero
	ldir				; Clear the data area
	dec (iy+7)			; Mark sd_sect invalid (FF000000)

do_interrupt:
	; Resume here to begin command
	xor a
	out (0x01),a

idle:
	ld e,0
idle_err:				; Errcode in E
	ei				; Allow interrupts
	jr setup_command

	defs 0x0038-$
interrupt_entry:			; Address 0038h
	ld sp,stack			; Reset stack
	ld iy,sysinfo			; Restore IY
	jr do_interrupt

	; Prepare to receive command
setup_command:
	xor a
	out (0x00),a
	out (0x02),a
	push iy				; Address for input
	pop hl
	ld a,l
	out (0x04),a
	ld a,h
	out (0x05),a
	ld a,4
	out (0x02),a
	ld a,e
	or 0x81				; Ready for command
	out (0x00),a
wait_for_cmd:
	in a,(0x00)			; DMA status
	rrca
	jr c,wait_for_cmd
	jr have_command

	defs 0x0066-$
nmi_entry:				; Address 0066h
	di
	jp nmi

have_command:
	; We have received an entire command sequence

	; Do we need to initialize the card?  We do this here, rather
	; than at reset time, because ABC80 is pretty generous with bus
	; and controller resets, and wants response almost immediately
	; after.
	bit 0,(iy+0x0e)
	call z,sd_init

	;
	; First, do common decoding operations
	;
	ld a,(iy+1)
	ld d,rambuf/(4*256)
	rlca				; 2 high bits of A -> 2 low bits of D
	rl d
	rlca
	rl d
	ld (iy+13),d			; Save pointer to addressed controller buffer

	; Reset auxilliary status
	xor a
	out (0x01),a

do_command:
	;
	; READ SECTOR command
	;
	bit 0,(iy+0)
	jr z,not_read_sector

	call load_buffer		; Load the SD buffer with the addressed SD sector
	jr c,data_error

	; Copy relevant half of SD buffer to controller buffer
	call bufsetup
	ldir				; Copy into controller buffer

not_read_sector:

	;
	; SECTOR TO HOST command
	;
	bit 1,(iy+0)			; SECTOR TO HOST
	jr z,not_sector_host
	ld c,0x07			; Controller->CPU
	call transfer
not_sector_host:

	;
	; SECTOR TO CONTROLLER command
	;
	bit 2,(iy+0)			; SECTOR TO CONTROLLER
	jr z,not_sector_controller
	ld c,0x05			; CPU->controller
	call transfer
not_sector_controller:

	;:
	; WRITE SECTOR command
	;:
	bit 3,(iy+0)
	jr z,not_write_sector

	; First, load the appropriate SD sector
	call load_buffer
	jr c,data_error

	; Copy controller buffer to relevant half of SD buffer
	call bufsetup
	ex de,hl			; Controller buffer -> SD buffer
	ldir				; Copy into controller buffer

	; Finally, write back buffer to SD card
	call write_buffer
	jr c,data_error

not_write_sector:

	;
	; All other commands ignored
	;
	jp idle

	;
	; Place to jump when an error is detected; auxilliary error
	; code should be placed in register E
	;
data_error:
	ld a,e				; Auxilliary status code
	out (0x01),a
	ld e,0x08			; ERROR
	jp idle_err

	;:
	; Set up register for a SD buffer -> controller buffer
	; transfer for the currently addressed buffer and sector.
	; For the opposite direction transfer, do ex de,hl before
	; actually transferring data
	;
bufsetup:
	ld hl,sd_buf/2
	ld bc,256			; 256 bytes
	ld a,(iy+3)			; Sector 
	rrca
	rl h				; HL -> proper half of SD buffer
	ld de,(abc_buf)			; Controller buffer
	ret
	
	;:
	; Transfer a sector to or from the controller.  Register C
	; should be set to 0x05 for host->controller and 0x07 for
	; controller->host
	;
transfer:
	xor a				; Controller buffer always aligned
	out (0x04),a
	ld a,(iy+13)			; High byte of addressed controller buffer
	out (c),a
	xor a
	out (0x03),a			; 256 bytes
	inc a				; A = 01h (READY)
	out (0x00),a
tr_wait:
	in a,(0x00)			; DMA engine status
	rrca				; Bit 0 = DMA done
	ret nc
	jr tr_wait
	
;:
; This routine converts the command contents to a SD sector number,
; and loads it if it is not currently loaded.
;
load_buffer:
	; Initialize buffer addresses
	xor a
	ld (iy+10),a
	ld (iy+11),a

	; Get track
	ld e,a				; Seek error
	ld a,(iy+2)
	cp 238				; 238 tracks/volume
	ccf
	ret c				; Bad sector 
	srl a
	ld (iy+9),a

	; Get sector
	ld a,(iy+3)
	rr a				; Convert to SD sector + low byte of track
	ld (iy+8),a

	; Compute unit number -> third LBA byte
	ld a,(iy+1)			; K1 = unit  + more
	and 0x07			; Unit 
	jr z,lb_haveunit		; Unit 0 skips loop
	; Each unit has 238 tracks, that is the maximum supported.
	; We multiply by 119 (238/2), this starts LBA bytes 2-3
	ld b,a
lb_getunit:
	ld a,(iy+9)
	add a,119
	ld (iy+9),a
	ld a,(iy+10)
	adc a,0
	ld (iy+10),a
	djnz lb_getunit
lb_haveunit:

	; Now sd_need contains the LBA sector (including upper 4 control bits)
	; that we need; sd_sect contains what's currently loaded in the SD
	; sector buffer.
	ld a,(iy+8)
	cp (iy+4)
	jr nz,lb_nomatch
	ld a,(iy+9)
	cp (iy+5)
	jr nz,lb_nomatch
	ld a,(iy+10)
	cp (iy+6)
	jr nz,lb_nomatch
	ld a,(iy+11)
	cp (iy+7)
	ret z				; If match, we're done

lb_nomatch:
	; No such luck.  Need to actually load the sector.
	ld hl,sd_need
	call fmt_lba			; Convert to byte count if appropriate
	di				; Don't confuse the card...
	ld a,17				; CMD17 - READ_SINGLE_BLOCK
	rst 8
	and a
	jr nz,bad_card

lb_wait_data:
	in a,(0x34)
	inc a
	jr z,lb_wait_data		; -1 = no response
	inc a
	jr nz,lb_error			; -2 = data token; other is error

	ld hl,sd_buf
	ld bc,0ff30h			; 255 bytes from port 30h
	inir
	inir				; ... plus 256 bytes

	; At this point, the last byte is latched, and the CRC16 is valid

	in a,(0x22)			; CRC16 MSB
	ld d,a
	in a,(0x23)			; CRC16 LSB
	ld e,a

	ini				; Store the last data byte

	in a,(0x30)			; CRC16 MSB
	xor d
	ld d,a
	in a,(0x30)			; CRC16 LSB
	ei
	xor e
	or d
	jr nz,lb_error

	ret				; F.C = 0 from "or d"

lb_error:
	; Read CRC error
	scf
	ld e,0x08			; Read error
	ret

bad_card:
	jp nmi

;:
; Write the buffer currently in the SD sector buffer back to the SD card
;
write_buffer:
	in a,(0x00)
	and 08h				; Write Protect?
	ld e,040h			; Write Protect error
	jr nz,wb_error
	ld hl,sd_sect
	call fmt_lba
	di				; Don't confuse the card...
	ld a,24				; CMD24 - WRITE_BLOCK
	rst 8
	and a
	jr nz,bad_card

	ld a,0FEh			; Data token
	out (0x3d),a
	ld bc,0x3d			; 256 bytes from port 0x3d
	ld hl,sd_buf
	outi				; Send first byte and clear CRC regs
	ld c,0x39
	otir
	otir

	in a,(0x23)			; CRC16 LSB
	ld c,a
	in a,(0x22)			; CRC16 MSB
	out (0x39),a
	ld a,c
	out (0x39),a

	in a,(0x30)			; Clock out the response token
	in a,(0x20)
	ld c,a				; Save response flag

wb_wait_unbusy:
	in a,(0x30)
	inc a
	jr nz,wb_wait_unbusy

	ei				; Transaction complete
	ld a,c
	and 1Fh
	cp 05h
	ret z				; If response valid, we're done

	ld e,20h			; Head load/write fault
wb_error:
	scf
	ret

;
; Format an LBA (in (HL)) appropriately, outputting to sect_arg.
; If we are in SDHC mode, just leave unchanged; if we are not,
; convert to a byte offset.
;
; Return with HL = sect_arg.
;
fmt_lba:
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)

	ld hl,sect_arg+3

	bit 6,(iy+0x17)
	jr nz,fmt_lba_sdhc

	; Format the LBA for non-SDHC mode
	sla c
	rl b
	rl e
	ld d,e
	ld e,b
	ld b,c
	ld c,0

	; Store the formatted LBA
fmt_lba_sdhc:
	ld (hl),d
	dec l
	ld (hl),e
	dec l
	ld (hl),b
	dec l
	ld (hl),c
	ret

;:
; NMI handing -- we receive an NMI when the SD card is removed.  Allow
; a DMA operation in process to complete, then report CMD/ERROR/READY
; with the error code of DRIVE NOT READY (80h) until the card is reinserted.
;:
; We do a full initialization after a card has been inserted; this includes
; invalidating the SD sector buffer!
;
nmi:
nmi_wait_dma:
	di
	xor a
	out (0x82),a			; Disable DMA

	ld a,0x08			; Drive not ready
	out (0x81),a			; Auxilliary status
	xor a
	out (0x80),a			; Primary status

nmi_wait_card:
	in a,(0x80)			; DMA/SD aux status
	bit 2,a				; Card present
	jr z,nmi_wait_card

	ld hl,0
	push hl
	retn				; Jump to reset point and enable NMI

;
; Initialize the SD card after power on or insert
;
sd_init:
	ld hl,init_arg			; Argument buffer
	;
	; Allow card to start up
	;
	ld a,3				; Low speed, CS# high
	out (0x22),a
	ld b,0				; 256 loops
	ld a,0xFF
sd_init_wait:
	out (0x31),a			; Send FF
	djnz sd_init_wait
	;
	; Reset card into SPI mode
	;
	ld a,1
	out (0x22),a			; Low speed, CS# low
	; command is already zero
	xor a
	rst 8				; CMD0 - GO_IDLE_STATE
	dec a
	jr nz,sd_init			; Not right?  Try again
	;
	; Enable command CRC checking, if available.
	;
	out (0x22),a			; A = 0: High speed, CS# low
	inc (hl)			; Enable CRC flag
	ld a,59				; CMD59 - CDC_ON_OFF
	rst 8
	; If this command fails, ignore it.
	dec (hl)			; Clear init_arg
	;
	; Probe for extended features
	;
	push hl
	ld hl,cmd8_arg
	ld a,8				; CMD8 - SEND_IF_COND
	rst 8
	pop hl
	bit 2,a
	jr nz,old_card			; CMD8 not supported
	dec a
	jr nz,sd_carderr		; Other error
	in a,(0x30)			; Command version
	in a,(0x30)			; Reserved bits
	in a,(0x30)			; Voltage range accepted
	ld b,a
	in a,(0x30)			; Check pattern
	cp 0xaa
	jr nz,sd_carderr
	bit 0,b				; 3300 mV supported
	jr z,sd_carderr
	;
	; Initialize card
	;
	set 6,(iy+0x1f)			; Bit 30 = SDHC permitted
old_card:
sd_init_acmd41:
	ld a,41
	call sd_acmd
	bit 2,a
	jr nz,sd_init_cmd1
	and a
	jr z,sd_probe_ocr
	dec a
	jr z,sd_init_acmd41
	jr sd_carderr

sd_init_cmd1:
	res 6,(iy+0x1f)			; Restore the zero buffer if needed
	inc a
	rst 8				; CMD1 - SEND_OP_COND
	and a
	jr z,sd_set_blklen
	dec a
	jr z,sd_init_cmd1
	jr sd_carderr

	;
	; Probe the OCR for SHDC mode
	;
sd_probe_ocr:
	res 6,(iy+0x17)			; Assume no SDHC
	bit 6,(iy+0x1f)			; Did we not even offer SDHC?
	jr z,no_sdhc
	res 6,(iy+0x1f)			; Clear init_arg
	ld a,58				; CMD58 - GET_OCR
	rst 8
	and a
	jr nz,sd_carderr
	push hl
	ld hl,card_ocr+3
	ld bc,0x0430			; 4 bytes from port 30h
	otdr
	pop hl
no_sdhc:

	;
	; Set block length to 512 bytes (2-4 GB cards sometimes
	; boot up defaulting to a longer block size, even though
	; that violates the spec.)
	;
sd_set_blklen:
	ld (iy+0x1d),2			; Argument buffer <- 512
	ld a,16				; CMD16 - SET_BLOCKLEN
	rst 8
	and a
	jr nz,sd_carderr
	ld (iy+0x1d),a			; Clear init_arg
sd_done:
	set 0,(iy+0x0e)			; Mark initialization done
	ret

;
; We end up here if we got an error during initialization.
; Room for better error handling...
;
sd_carderr:
	jp sd_init

;
; Issue a SD card application command and obtain the response.
; The command number is in register A, and the 32-bit argument
; is in the memory buffer pointed to by HL in littleendian order.
; Returns the (first byte of) the response code in A, all other
; registers are preserved.
;
no_acmd:
	pop hl				; Drop AF on stack
	pop hl
	ret				; Return with unknown command error

sd_acmd:
	push hl
	push af
	ld a,55				; CMD55 - APP_CMD
	ld hl,zero_arg
	rst 8
	bit 2,a				; Unknown command (very old card)?
	jr nz,no_acmd
	and 0FEh
	jp nz,nmi			; APP_CMD shouldn't error...
	pop af
	pop hl
	; Fall through

;
; Issue an SD card command and obtain the response.
; The command number is in register A, and the 32-bit argument
; is in the memory buffer pointed to by HL in littleendian order.
; Returns the (first byte of) the response code in A, all other
; registers are preserved.
;
; Can be invoked with RST 8.
;
sd_cmd:
	push bc
	or 0x40				; Start bit
	out (0x3d),a			; Send command code
	inc hl				; HL -> end of buffer
	inc hl
	inc hl
	ld bc,0x0439
	otdr				; Send argument (MSB->LSB)
	in a,(0x29)			; Get CRC7
	out (0x39),a			; Send CRC7
	ld b,16
sd_cmd_resp:
	in a,(0x30)
	and a
	jp p,sd_cmd_out
	djnz sd_cmd_resp
	jp nmi				; Card not responding, assume removed
sd_cmd_out:
	inc hl				; Restore HL
	pop bc
	ret

;
; Read-only initialized data
;
cmd8_arg:				; Check pattern, 3.3 V
	defb 0xaa, 0x01, 0x00, 0x00
